/*
    Copyright (c) 2011,2012 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>
            Michael Troß, aizo GmbH <michael.tross@aizo.com>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __DATABASE_H__
#define __DATABASE_H__

#include <boost/thread/mutex.hpp>
#include <boost/shared_ptr.hpp>
#include <map>
#include <string>
#include <list>

#include "sqlite3_wrapper.h"
#include "dsmapi-wrapper.h"
#include "tasks.h"
#include "hwstats.h"

#define DATABASE_MAX_FILESIZE (2 * 1024 * 1024)

#define TRANSMISSION_QUALITY_UPSTREAM_MAX   62 // best
#define TRANSMISSION_QUALITY_DOWNSTREAM_MAX 6  // worst

namespace dSA
{

class Database
{
public:
    typedef std::pair<std::string, int> field;
    typedef std::map<std::string, boost::shared_ptr<field> > result_data;

    typedef struct
    {
        boost::shared_ptr<dSLongTask> task;
        boost::shared_ptr<result_data> data;
    } task_result_summary;

    typedef struct
    {
        std::string dsid;
        int upstream;
        int downstream;
        time_t timestamp;
        int errorcode;
    } transmission_quality_device_record;

    typedef struct
    {
        std::string dsid;
        std::string dsmdid;
        int device_id;
        int resultcode;
        time_t timestamp;
    } device_presence_record;

    Database();
    static Database *createInstance();
    static Database *getInstance();
    ~Database();

    /// \brief Return size of database file on disk
    size_t getFileSize() const;

    /// \brief Shrink database file size by running the VACUUM; command
    void executeVacuum() const;

    /// \brief Retrieve database version information, if database was
    /// migrated, the "old" parameter will contain the original version.
    void getVersionInfo(int *current, int *old) const;

    /// \brief Retrieve a value from the configuration dable
    std::string getConfig(std::string key);

    /// \brief Add/Set value in the configuration table
    void setConfig(std::string key, std::string value);


    /// \brief Saves a task in the database
    /// \parameter task task that should be saved
    /// \return task ID
    long long int saveTask(boost::shared_ptr<const dSLongTask> task);

    /// \brief Removes a task given by id from the database
    void removeTask(long long int task_id);

    /// \brief Remove tasks given by status from the database
    void removeTasksByStatus(task_status status);

    /// \brief Update the status of a task in the databse
    ///
    /// This function will automatically take care of satting the relevant
    /// timestamps.
    void updateTaskStatus(long long int task_id, task_status status);

    /// \brief Set new status on all tasks with given old status
    void updateTaskStatus(task_status old_status, task_status new_status);

    /// \brief Load task type information
    task_type getTaskType(long long int task_id);

    /// \brief Load task from the database given by id
    boost::shared_ptr<dSLongTask> getTask(long long int task_id);

    /// \brief Retrieves a list of tasks that have a certain state
    boost::shared_ptr<std::list<boost::shared_ptr<dSLongTask> > >
                                          getTasksByStatus(task_status status);

    /// \brief Retrieve a name for a statistic field
    std::pair<std::string,std::string> getStatisticDescription(statistic_field_t index);

    /// \brief Retrieve full list of statistic field short names (mode 0)
    /// or description (mode 1)
    boost::shared_ptr<std::list<std::pair<statistic_field_t,std::string> > >
        getStatisticTableNames(int mode = 0);

    /// \brief Save result of a device transmission quality test
    void saveTransmissionQuality(long long int task_id, std::string device_dsid,
                                 uint16_t upstream, uint8_t downstream,
                                 int error_code);

    /// \brief Save result of device presence checks
    ///
    /// The resultcode may be one of the following:
    ///       1 unknown      presence check aborted before schedule
    ///       2 scheduled    presence check not executed for this device
    ///       3 sent         query sent out but no response received
    ///       4 replied      successful device reply
    ///       5 removed      device state changed to inactive due to missing response
    ///       6 discovered   device state changed to active
    void saveDevicePresence(long long int task_id,
            std::string device_dsid, std::string dsm_dsid, uint16_t device_id,
            uint8_t resultcode);
    void updateDevicePresenceStatus(long long int task_id,
            std::string dsm_dsid, uint16_t device_id, uint8_t resultcode);

    /// \brief Save result of device statistic query
    void saveStatisticDevice(long long int task_id,
            dsid_t device_dsid, dsid_t dsm_dsid, uint16_t device_id,
            statistic_field_t index, long long int value,
            uint8_t resultcode);

    /// \brief Save result of device statistic query
    void saveStatisticDevice(long long int task_id,
            dsid_t device_dsid, dsid_t dsm_dsid, uint16_t device_id,
            boost::shared_ptr<std::list<boost::shared_ptr<StatisticValue> > > svalues);

    /// \brief Save result of meter statistic query
    void saveStatisticMeter(long long int task_id, dsid_t dsm_dsid,
            boost::shared_ptr<std::list<boost::shared_ptr<StatisticValue> > > svalues);

    /// \brief Writes the list of CBData blocks into the table
    void saveUpstreamMetric(boost::shared_ptr<std::list<
            boost::shared_ptr<dSMApiSelectiveListener::CBData> > > cbdatalist);

    /// \brief Cleanup upstream table
    void removeUpstreamMetricByTime(time_t threshold);


    /// \brief Returns the summarized version of task data.
    ///
    /// The data and the keys will vary depending on the task
    boost::shared_ptr<task_result_summary>
                                    getTaskResultSummary(long long int task_id);

    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
                                        loadTasks(long long int limit,
                                                  long long int offset);

    long long int transmissionQualityResultCount(long long int task_id);

    boost::shared_ptr<std::list<
        boost::shared_ptr<const transmission_quality_device_record> > >
            getTransmissionQualityResults(long long int task_id,
                                          long long int limit,
                                          long long int offset);

    boost::shared_ptr<std::list<
        boost::shared_ptr<const device_presence_record> > >
            getDevicePresenceResults(long long int task_id,
                    long long int limit,
                    long long int offset);

    void saveDSM(boost::shared_ptr<dSM> dsm);
    boost::shared_ptr<dSM> loadDSM(dsid_t dsid);
    boost::shared_ptr<std::list<dsid_t> > getDSMdSIDs();
    void deleteDSM(dsid_t dsid);
    void updateDSMPresence(bool presence);

    void saveDevice(boost::shared_ptr<Device> device);
    boost::shared_ptr<Device> loadDevice(dsid_t dsid);
    boost::shared_ptr<Device> loadDevice(dsid_t dsid, dsid_t dsm_dsid);
    boost::shared_ptr<Device> loadDevice(dsid_t dsid, uint16_t device_id);
    void deleteDevice(dsid_t dsid);
    void deleteDevice(dsid_t dsid, dsid_t dsm_dsid);
    void deleteDevice(dsid_t dsid, uint16_t device_id);
    void deleteDevice(dsid_t dsid, dsid_t dsm_dsid, uint16_t device_id);

    boost::shared_ptr<std::list<boost::shared_ptr<Device> > > loadDevices(
            dsid_t dsm_dsid, bool only_active);
    boost::shared_ptr<std::list<boost::shared_ptr<Device> > > loadDevices(
            bool only_active);

    boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                loadDeviceList(dsid_t dsid, bool onlyActive = true);
    boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                loadDeviceListM(dsid_t dsm_dsid, uint16_t device_id);

    void updatePresence(bool presence);

private:
    static Database* m_instance;
    SQLite3 *m_sqlite3;
    boost::mutex m_mutex;
    int m_old_version;
    int m_version;

    boost::shared_ptr<dSLongTask> createTaskFromRow(
                                    boost::shared_ptr<SQLite3::row_result> rr);
    boost::shared_ptr<field> singleValueFromQuery(
                                    boost::shared_ptr<SQLite3::query_result> q);
    boost::shared_ptr<transmission_quality_device_record>
                  TQDevRecordFromRow(boost::shared_ptr<SQLite3::row_result> rr);
    boost::shared_ptr<dSM> createDSMFromRow(dsid_t dsid,
                                    boost::shared_ptr<SQLite3::row_result> rr);
    std::string deviceKey(dsid_t dsid, dsid_t dsm_dsid);
    boost::shared_ptr<Device> createDeviceFromRow(
                                    boost::shared_ptr<SQLite3::row_result> rr);
};

} // namespace

#endif//__DATABASE_H__
